/*
   tsbdecode - decode programs from an HP 2000 TSB file (from tsbextract)

   Copyright 1999, 2000, 2002, 2003 Eric Smith

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as published
   by the Free Software Foundation.  Note that permission is not granted
   to redistribute this program under the terms of any other version of the
   General Public License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 * Expects the start of the file to contain the directory entry, which
 * is normal for files extracted from a hibernate tape by tsbextract.
 *
 * TODO:
 *   test extended string representation 
 *     (p. 5-34 in internals manual, p. 219 in PDF)
 *   computed GOTO/GOSUB
 *
 *   compiled programs
 */

/*
 * Fixed (Pete Turnbull, November 2003):
 *   floating point
 *     values less than 1
 *     leading and trailing zeros
 *   PRINT USING parameters
 *   cosmetic spacing (to match old listings)
 */
 
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>


void print_char (FILE *out, unsigned char c)
{
  /* Danger, Will Robinson!
     If you don't output "\" as "\\", then "\012" is ambiguous
     If you do, the listing is wrong :-(  */
     
  /* if (c == '\\')
    fprintf (out, "\\\\");
  else */ if ((c >= ' ') && (c <= '~'))
    fprintf (out, "%c", c);
  else
    fprintf (out, "\\%03o", c);
}


float decode_fp (uint32_t v)
{
  int mant_sign = (v & 0x80000000) != 0;
  int exp_sign = v & 1;
  float val = ((v & 0x7fffff00) >> 8) / (1.0 * (1 << 23));
  int exp = (v >> 1) & 0x7f;

  if (mant_sign)
    val = -val;

  if (exp_sign)
    val /= (1 << (128-exp));
  else
    val *= (1 << exp);

  return (val);
}


/* return number of characters printed */
int print_fp (FILE *out, uint32_t v)
{
  char value[32];
  float fval = decode_fp (v);
  sprintf(value, "%g", fval);
  return (fprintf (out, "%s", value[0]=='0' && value[1]=='.' ? value+1 : value));
}


#define MAX_FN_LEN 6

uint32_t get16 (uint8_t *buf, uint32_t offset)
{
  return ((buf [offset] & 0xff) << 8) | (buf [offset+1] & 0xff);
}

char * file_flag_names [16] =
{
  "unrestricted",
  "protected",
  "locked",
  "private",
  "unused",
  "unused",
  "unused",
  "unused",
  "unused",
  "unused",
  "unused",
  "fcp",
  "mwa",
  "pfa",
  "output",
  "input"
};


char *statement [64] =
{
  /* 00-03 */ NULL,       NULL,        NULL,       NULL,
  /* 04-07 */ NULL,       NULL,        NULL,       NULL,
  /* 10-13 */ NULL,       NULL,        NULL,       NULL,
  /* 14-17 */ NULL,       NULL,        NULL,       NULL,
  /* 20-23 */ NULL,       NULL,        NULL,       NULL,
  /* 24-27 */ NULL,       NULL,        NULL,       NULL,
  /* 30-33 */ NULL,       NULL,        "SYSTEM ",  "CONVERT ",
  /* 34-37 */ "LOCK ",    "UNLOCK ",   "CREATE ",  "PURGE ",
  /* 40-43 */ "ADVANCE ", "UPDATE ",   "ASSIGN ",  "LINPUT ",
  /* 44-47 */ "IMAGE ",   "COM ",      "LET ",     "DIM ",
  /* 50-53 */ "DEF ",     "REM",       "GOTO ",    "IF ",
  /* 54-57 */ "FOR ",     "NEXT ",     "GOSUB ",   "RETURN",
  /* 60-63 */ "END",      "STOP",      "DATA ",    "INPUT ",
  /* 64-67 */ "READ ",    "PRINT ",    "RESTORE ", "MAT ",
  /* 70-73 */ "FILES ",   "CHAIN ",    "ENTER ",   "" /* "let " */,
  /* 74-77 */ NULL,       NULL,        NULL,       NULL
};

char *operator [64] =
{
  /* 00 */ "",  /* end of formula */
  /* 01 */ "\"",
  /* 02 */ ",",
  /* 03 */ ";",
  /* 04 */ "#",   /* file */
  /* 05-07 */ NULL, NULL, NULL,
  /* 10 */ ")",
  /* 11 */ "]",
  /* 12 */ "[",
  /* 13 */ "(",
  /* 14 */ "+",   /* unary */
  /* 15 */ "-",   /* unary */
  /* 16 */ ",",   /* subscript */
  /* 17 */ "=",   /* assignment */
  /* 20 */ "+",
  /* 21 */ "-",
  /* 22 */ "*",
  /* 23 */ "/",
  /* 24 */ "^",
  /* 25 */ ">",
  /* 26 */ "<",
  /* 27 */ "#",
  /* 30 */ "=",   /* equal */
  /* 31 */ NULL,
  /* 32 */ " AND ",
  /* 33 */ " OR ",
  /* 34 */ " MIN ",
  /* 35 */ " MAX ",
  /* 36 */ " <> ",
  /* 37 */ " >= ",
  /* 40 */ " <= ",
  /* 41 */ " NOT ",
  /* 42 */ "**",  /* power */
  /* 43 */ " USING ",
  /* 44 */ "PR",
  /* 45 */ "WR",
  /* 46 */ "NR",
  /* 47 */ "ERROR ",
  /* 50-57 */ NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /* 60 */ "END ",
  /* 61-62 */ NULL, NULL,
  /* 63 */ "INPUT ",
  /* 64 */ "READ ",
  /* 65 */ "PRINT ",
  /* 66-73 */ NULL, NULL, NULL, NULL, NULL, NULL,
  /* 74 */ " OF ",
  /* 75 */ " THEN ",
  /* 76 */ " TO ",
  /* 77 */ " STEP "
};

char *function [32] =
{
  /* 00 */ "CTL",
  /* 01 */ "TAB",
  /* 02 */ "LIN",
  /* 03 */ "SPA",
  /* 04 */ "TAN",
  /* 05 */ "ATN",
  /* 06 */ "EXP",
  /* 07 */ "LOG",
  /* 10 */ "ABS",
  /* 11 */ "SQR",
  /* 12 */ "INT",
  /* 13 */ "RND",
  /* 14 */ "SGN",
  /* 15 */ "LEN",
  /* 16 */ "TYP",
  /* 17 */ "TIM",
  /* 20 */ "SIN",
  /* 21 */ "COS",
  /* 22 */ "BRK",
  /* 23 */ "ITM",
  /* 24 */ "REC",
  /* 25 */ "NUM",
  /* 26 */ "POS",
  /* 27 */ "CHR$",
  /* 30 */ "U0S$",  /* ??? */
  /* 31 */ "SYS",
  /* 32 */ "<predef32>",
  /* 33 */ "ZER",
  /* 34 */ "CON",
  /* 35 */ "IDN",
  /* 36 */ "INV",
  /* 37 */ "TRN",
};


void print_operator (FILE *out,
		     uint8_t *buf,
		     uint32_t p,   /* start of statement (length word) */
		     uint32_t *q)  /* current loc */
{
  int first = (*q == (p+4));
  uint16_t word;
  int msb, op, name, operand;
  char *opname;
  uint32_t val;
  uint32_t word_count = get16 (buf, p + 2);
  int st = (get16 (buf, p + 4) >> 9) & 077;
  int count;
  /* static int dimension; */

  word = get16 (buf, *q);

  msb = word >> 15;
  op = (word >> 9) & 077;
  name = (word >> 4) & 037;
  operand = word & 017;

#if 0
  fprintf (out, "<%o-%02o-%02o-%02o>", msb, op, name, operand);
#endif

  if (first)
    opname = statement [op];
  else
    opname = operator [op];

  (*q) += 2;

  if (opname)
    fprintf (out, "%s", opname);
  else
    fprintf (out, "?%02o ", op);

  if ((first && (op == 044)) ||    /* IMAGE statement */
      ((! first) && (op == 001)))  /* string constant */
    {
      int len;
      if (word & 0x8100)  /* should never happen */
	fprintf (out, "?");
      len = word & 0xff;
      while (len > 0)
	{
	  word = get16 (buf, *q);
	  (*q) += 2;
	  print_char (out, word >> 8);
	  if (len > 1)
	    print_char (out, word & 0xff);
	  len -= 2;
	}
      if ((! first) && (op == 001))
	fprintf (out, "\"");
    }
  else if (! msb)
    {
      if ((name >= 033) && (name <= 036))
	{
	  /* extended string variable */
	  fprintf (out, "%c%d$", 'A' + 16 * ((name - 1) % 2) + operand - 1,
		   (name - 033) / 2);
	}
      else switch (operand)
	{
	case 0:  /* string variable or null operand */
	  if (name)
	    fprintf (out, "%c$", 'A' + name - 1);
	  break;
	case 1:  /* one-dimensional array */
	case 2:  /* two-dimensional array */
	case 3:  /* array of unknown dimensionality */
	  /* dimension = operand; */
	  fprintf (out, "%c", 'A' + name - 1);
	  break;
	case 4:  /* scalar with no digit suffix */
	  fprintf (out, "%c", 'A' + name - 1);
	  break;
	case 017:  /* user-defined function */
	  fprintf (out, "FN%c", 'A' + name - 1);
	  break;
	default:  /* 005..016:  scalar with digit suffix */
	  fprintf (out, "%c%d", 'A' + name - 1, operand - 5);
	  break;
	}
    }
  else
    {
      switch (operand)
	{
	case 000:  /* numerical constant */
	  val = (get16 (buf, *q) << 16) + get16 (buf, 2 + *q);
	  print_fp (out, val);
	  (*q) += 4;
	  break;
	case 001:
	case 002:
	  fprintf (out, "<unknown operand type %d>", operand);
	  break;
	case 003:  
	  if ((st == 045) ||	/* COM */
	      (st == 047) ||	/* DIM: formal dimension */
	      (op == 043))  	/* USING */
	    count = 1;
	  else /* target line number or list (GOTO/GOSUB <expr> OF <list>) */
	    count = (p + word_count * 2 - (*q)) / 2;
	  while (count--)
	    {
	      fprintf (out, "%d", get16 (buf, *q));
	      if (count)
		fprintf (out, ",");
	      (*q) += 2;
	    }
	  break;
	case 017:  /* predefined function */
	  fprintf (out, "%s", function [name]);
	  break;
	case 4:  /* user-defined function parameter with no digit suffix */
	  fprintf (out, "%c", 'A' + name - 1);
	  break;
	default:  /* 005..016:  user-defined fn parameter with digit suffix */
	  fprintf (out, "%c%d", 'A' + name - 1, operand - 5);
	  break;
	}
    }
}


/* disk files: */
#define HP_2000_PROGRAM      0
#define HP_2000_SEMICOMPILED 1
#define HP_2000_FILE         4
#define HP_2000_ASCII        6
/* tape file only: */
#define HP_2000_HIBERNATE    8

int file_type;


int quiet;


void print_date (FILE *out, uint32_t date)
{
  fprintf (out, "%d/%d", 1900 + (date >> 9), date & 0x1ff);
}

void dump_directory_entry (FILE *out, uint8_t *buf)
{
  uint32_t uid;
  uint32_t flags;
  char uids [6];
  char filename [MAX_FN_LEN + 1];
  int i;

  uid = get16 (buf, 0);
  if (uid == 0xffff)
    {
      if (! quiet)
	fprintf (out, "End of directory\n");
      return;
    }
  uids [0] = (uid / 1024) + '@';
  sprintf (& uids [1], "%03d", uid & 0x3ff);
  for (i = 0; i < MAX_FN_LEN; i++)
    filename [i] = buf [i+2] & 0x7f;
  filename [sizeof (filename) - 1] = '\0';
  if (! quiet)
    {
      fprintf (out, "ID %s filename '%s'", uids, filename);
      if (buf [2] & 0x80)
	fprintf (out, " ASCII");
      if (buf [4] & 0x80)
	fprintf (out, " file");
      if (buf [6] & 0x80)
	fprintf (out, " semi-compiled");
    }
  file_type = (((buf [4] & 0x80) >> 5) +
	       ((buf [2] & 0x80) >> 6) +
	       ((buf [6] & 0x80) >> 7));
  flags = get16 (buf, 14);
  if (! quiet)
    {
      for (i = 15; i >= 0; i--)
	{
	  if (flags & (1 << i))
	    fprintf (out, " %s", file_flag_names [i]);
	}
      fprintf (out, ", last referenced ");
      print_date (out, get16 (buf, 10));
      fprintf (out, ", last modified %d", get16 (buf, 12));
  
      if (buf [4] & 0x80)
	fprintf (out, ", length %d records of %d words", get16 (buf, 22),
		 get16 (buf, 8));
      else
	{
	  fprintf (out, ", length %d words", 65536 - get16 (buf, 22));
	  fprintf (out, ", start loc %06o", get16 (buf, 8));
	}
      fprintf (out, "\n");
    }
}



void dump_statement_hex (FILE *out, uint8_t *buf, uint32_t p)
{
  uint32_t word_count = get16 (buf, p+2);
  int i;
  uint32_t word;

  for (i = 0; i < word_count; i++)
    {
      word = get16 (buf, p + i*2);
      fprintf (out, "%02x %02x ", (word >> 8) & 0xff, word & 0xff);
    }
  fprintf (out, "\n");
}

void dump_statement (FILE *out, uint8_t *buf, uint32_t p)
{
  uint32_t line_number = get16 (buf, p);
  uint32_t word_count = get16 (buf, p + 2);
  uint32_t word = get16 (buf, p + 4);
  int st = (word >> 9) & 077;
  uint32_t q = p + 4;

  fprintf (out, "%d ", line_number);

  if ((st == 051) || (st == 070))  /* REM or FILES */
    {
      fprintf (out, statement [st]);
      if (word & 0xff)
	print_char (out, word & 0xff);
      q += 2;
      word_count -= 3;
      while (word_count--)
	{
	  word = get16 (buf, q);
	  print_char (out, word >> 8);
	  if ((word_count != 0) || (word & 0xff))
	    print_char (out, word & 0xff);
	  q += 2;
	}
    }
  else
    {
      while (q < (p + 2 * word_count))
	print_operator (out, buf, p, & q);
    }
  fprintf (out, "\n");
}


void dump_program (FILE *f, FILE *out)
{
  int file_length;
  int e;
  uint32_t p = 0;
  uint32_t prev_line = 0;
  uint32_t line_number;
  uint32_t word_count;
  char buf [20480];

  file_length = fread (buf, 1, 20480, f);
  fprintf (out, "file length %d bytes\n", file_length);
  e = ferror (f);
  if (e)
    {
      fprintf (stderr, "error reading program\n");
      return;
    }

  while (p < file_length)
    {
      if ((p + 4) > file_length)
	{
	  fprintf (out, "corrupt?  not enough space left for a line number and length\n");
	  break;
	}
      line_number = get16 (buf, p);
      if (line_number < prev_line)
	{
	  fprintf (out, "corrupt?  line number out of sequence\n");
	}
      word_count = get16 (buf, p + 2);
      if ((line_number == 0) && (word_count == 0))
	break;
      if (word_count < 2)
	{
	  fprintf (out, "corrupt?  word count = %d\n", word_count);
	  break;
	}
      if ((p + word_count * 2) > file_length)
	{
	  fprintf (out, "corrupt, program line extends past end of file!\n");
	  break;
	}
#if 0
      dump_statement_hex (out, buf, p);
#endif
      dump_statement (out, buf, p);
      p += word_count * 2;
    }
}


#define FILE_BLOCK_SIZE 512


void dump_record (FILE *out, uint8_t *buf)
{
  int pos = 0;
  int i = 0;
  int len;

  while (i < FILE_BLOCK_SIZE)
    {
      if (buf [i] == 0x02)
	{
	  len = buf [i+1];
	  i += 2;
	  if (i + len > FILE_BLOCK_SIZE)
	    {
	      fprintf (out, "string too long for record\n");
	      return;
	    }
	  if (pos)
	    {
	      fprintf (out, "\n");
	      pos = 0;
	    }
	  if (! quiet)
	    fprintf (out, "\"");
	  while (len > 0)
	    {
	      print_char (out, buf [i]);
	      if (len > 1)
		print_char (out, buf [i+1]);
	      i += 2;
	      len -= 2;
	    }
	  if (! quiet)
	    fprintf (out, "\"");
	  fprintf (out, "\n");
	  pos = 0;
	}
      else if ((buf [i] == 0xff) && (buf [i+1] == 0xfe))
	{
	  /* end of record */
	  if (pos)
	    {
	      fprintf (out, "\n");
	      pos = 0;
	    }
	  if (! quiet)
	    fprintf (out, "logical end of record\n");
	  return;
	}
      else
	{
	  /* numeric */
	  pos += print_fp (out, ((buf [i] << 24) | (buf [i+1] << 16) |
				 (buf [i+2] << 8) | buf [i+3]));
	  pos += fprintf (out, "  ");
	  i += 4;
	  if (pos >= 70)
	    {
	      fprintf (out, "\n");
	      pos = 0;
	    }
	}
    }
  if (pos)
    fprintf (out, "\n");
  if (! quiet)
    fprintf (out, "physical end of record\n");
}


void dump_file (FILE *f, FILE *out)
{
  int rec = 0;
  uint8_t buf [FILE_BLOCK_SIZE];

  for (;;)
    {
      if (1 != fread (buf, FILE_BLOCK_SIZE, 1, f))
	{
	  if (feof (f))
	    exit (0);
	  fprintf (stderr, "error reading file\n");
	  exit (2);
	}
      if (! quiet)
	fprintf (out, "record %d\n", rec++);
      dump_record (out, buf);
    }
}


int main (int argc, char *argv[])
{
  FILE *f;
  uint8_t header [24];

  quiet = 1;

  if (argc != 2)
    {
      fprintf (stderr, "usage: %s file\n", argv [0]);
      exit (1);
    }

  f = fopen (argv [1], "rb");
  if (! f)
    {
      fprintf (stderr, "error opening file\n");
      exit (2);
    }

#if 1
  if (1 != fread (header, sizeof (header), 1, f))
    {
      fprintf (stderr, "error reading file header\n");
      exit (2);
    }
    dump_directory_entry (stdout, header);
#endif

  switch (file_type)
    {
    case HP_2000_PROGRAM:
      dump_program (f, stdout);
      break;
    case HP_2000_SEMICOMPILED:
      fprintf (stderr, "Semi-compiled programs not yet supported\n");
      dump_program (f, stdout);
      break;
    case HP_2000_FILE:
      dump_file (f, stdout);
      break;
    case HP_2000_ASCII:
      fprintf (stderr, "ASCII files not yet supported\n");
      break;
    default:
      fprintf (stderr, "unknown file type\n");
    }

  fclose (f);
  exit (0);
}
