/*
   tsbextract - extract files and programs from an HP 2000 TSB hibernate tape

   Copyright 1999, 2000, 2002, 2003 Eric Smith

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as published
   by the Free Software Foundation.  Note that permission is not granted
   to redistribute this program under the terms of any other version of the
   General Public License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 * The start of each output file will contain a copy of the directory entry.
 */

#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "tapeio.h"

#define MAX_REC_LEN 32768


char *progname;

void print_usage (FILE *f)
{
  fprintf (f, "Usage: %s in\n", progname);
}

void fatal (int retval, char *fmt, ...)
{
  va_list ap;

  if (fmt)
    {
      fprintf (stderr, "%s: ", progname);
      va_start (ap, fmt);
      vfprintf (stderr, fmt, ap);
      va_end (ap);
    }

  if (retval == 1)
    print_usage (stderr);

  exit (retval);
}


#define BYTES_PER_LINE 16

void dump (FILE *f, uint8_t *buf, uint32_t len)
{
  uint32_t i, j;

  for (i = 0; i < len; i += BYTES_PER_LINE)
    {
      fprintf (f, "  %04x: ", i);
      for (j = 0; j < BYTES_PER_LINE; j++)
	if ((i + j) < len)
	  fprintf (f, "%02x ", 0xff & buf [i+j]);
	else
	  fprintf (f, "   ");
      for (j = 0; j < BYTES_PER_LINE; j++)
	if ((i + j) < len)
	  {
	    uint8_t c = buf [i+j] & 0x7f;
	    if ((c >= ' ') && (c <= '~'))
	      fprintf (f, "%c", c);
	    else
	      fprintf (f, ".");
	  }
	else
	  fprintf (f, " ");
      fprintf (f, "\n");
    }
}


#define UID_LEN 4
#define MAX_FN_LEN 6

uint32_t get16 (uint8_t *buf, uint32_t offset)
{
  return ((buf [offset] & 0xff) << 8) | (buf [offset+1] & 0xff);
}

char * file_flag_names [16] =
{
  "unrestricted",
  "protected",
  "locked",
  "private",
  "unused",
  "unused",
  "unused",
  "unused",
  "unused",
  "unused",
  "unused",
  "fcp",
  "mwa",
  "pfa",
  "output",
  "input"
};


/* disk files: */
#define HP_2000_PROGRAM      0
#define HP_2000_SEMICOMPILED 1
#define HP_2000_FILE         4
#define HP_2000_ASCII        6
/* tape file only: */
#define HP_2000_HIBERNATE    8

int hp_2000_file_type;


void print_hp_2000_date (FILE *f, uint32_t date)
{
  fprintf (f, "%d/%d", 1900 + (date >> 9), date & 0x1ff);
}


void dump_hp_2000_directory_entry (FILE *f, uint8_t *buf, char *dir, char *fn)
{
  uint32_t uid;
  uint32_t flags;
  int i;

  uid = get16 (buf, 0);
  if (uid == 0xffff)
    {
      fprintf (f, "End of directory\n");
      return;
    }

  sprintf (dir, "%c%03d", (uid / 1024) + '@', uid & 0x3ff);

  for (i = 0; i < MAX_FN_LEN; i++)
    {
      fn [i] = buf [i+2] & 0x7f;
      if (fn [i] == ' ')
	fn [i] = '\0';
    }
  fn [MAX_FN_LEN] = '\0';

  fprintf (f, "ID %s filename '%s'", dir, fn);
  if (buf [2] & 0x80)
    fprintf (f, " ASCII");
  if (buf [4] & 0x80)
    fprintf (f, " file");
  if (buf [6] & 0x80)
    fprintf (f, " semi-compiled");
  hp_2000_file_type = (((buf [4] & 0x80) >> 5) +
		       ((buf [2] & 0x80) >> 6) +
		       ((buf [6] & 0x80) >> 7));
  flags = get16 (buf, 14);
  for (i = 15; i >= 0; i--)
    {
      if (flags & (1 << i))
	fprintf (f, " %s", file_flag_names [i]);
    }
  fprintf (f, ", last referenced ");
  print_hp_2000_date (f, get16 (buf, 10));
  fprintf (f, ", last modified %d", get16 (buf, 12));
  
  if (buf [4] & 0x80)
    fprintf (f, ", length %d records of %d words", get16 (buf, 22),
	     get16 (buf, 8));
  else
    {
      fprintf (f, ", length %d words", 65536 - get16 (buf, 22));
      fprintf (f, ", start loc %06o", get16 (buf, 8));
    }
  fprintf (f, "\n");
}


int create_directory (char *dn)
{
  struct stat buf;

  if (stat (dn, & buf) == 0)
    {
      /* exists, is it a directory? */
      if (S_ISDIR(buf.st_mode))
	return (1);
      else
	return (0);
    }
  else
    {
      if (mkdir (dn, S_IRUSR | S_IWUSR | S_IXUSR |
		 S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) == 0)
	return (1);
      else
	return (0);
    }
}


FILE *dump_hp_2000_file_header (FILE *f, uint8_t *buf, uint32_t len)
{
  FILE *out;
  char fn [UID_LEN + MAX_FN_LEN + 2];

  if (len < 24)
    {
      fprintf (f, "file header too short for directory entry\n");
      return (NULL);
    }

  dump_hp_2000_directory_entry (f, buf, fn, fn + UID_LEN + 1);

#if 1
  if (! create_directory (fn))
    {
      fprintf (stderr, "error creating directory '%s'\n", fn);
      return (NULL);
    }
  fn [UID_LEN] = '/';
#else
  fn [UID_LEN] = '-';
#endif
  out = fopen (fn, "w");
  if (! out)
    {
      fprintf (stderr, "error opening file '%s'\n", fn);
      return (NULL);
    }

#if 1
  /* write the directory entry followed by the data */
  fwrite (buf, len, 1, out);
#else
  if (len > 24)
    fwrite (buf + 24, len - 24, 1, out);
#endif

  return (out);
}


void dump_hp_2000_hibernate (FILE *f, uint32_t record, uint8_t *buf, uint32_t length)
{
  /* $$$ need better way to track what kind of record we expect, since a lot
     of the stuff is of variable lengths */
  switch (record)
    {
    case 1:  printf ("Equipment table\n"); break;
    case 2:  printf ("Direc table\n"); break;
    case 3:  printf ("ID table\n"); break;
    }
}


uint8_t hp_2000_tape_label [6] = { 0xff, 0xf6, 'L', 'B', 'T', 'S' };

void dump_hp_2000_dump_tape (FILE *f, uint32_t file, uint32_t record, uint8_t *buf, uint32_t len)
{
  static FILE *out;
  if ((file == 0) && (record == 0) && (len == 20))
    {
      if (strncmp (buf, hp_2000_tape_label, sizeof (hp_2000_tape_label)) == 0)
	{
	  hp_2000_file_type = HP_2000_HIBERNATE;
	  printf ("HP 2000 HIBernate tape, reel %d\n", get16 (buf, 8));
	  printf ("year %d, hour %d, tenths of seconds %d\n", get16 (buf, 10),
		  get16 (buf, 12), get16 (buf, 14));
	  printf ("system level %d\n", get16 (buf, 16));
	  printf ("feature level %d\n", get16 (buf, 18));
	}
      else
	fprintf (stderr, "unrecognized tape label\n");
    }
  else if ((file == 0) && (record > 0) && (hp_2000_file_type == HP_2000_HIBERNATE))
    dump_hp_2000_hibernate (f, record, buf, len);
  else if (record == 0)
    {
      if (out)
	fclose (out);
      out = dump_hp_2000_file_header (f, buf, len);
    }
  else
    {
      if (out)
	fwrite (buf, len, 1, out);
    }
}


void dump_hp_2000_dump_tape_eof (FILE *f, uint32_t file, uint32_t record)
{
}


void dump_hp_2000_abs_record (FILE *f, uint8_t *buf, uint32_t len)
{
  uint32_t l;
  uint32_t checksum = 0;
  uint32_t tape_checksum;
  uint32_t i;

  l= buf [0] * 2 + 6;
  if ((l != len) || (buf [1] != 0))
    {
      fprintf (f, "first word of record is not length (expected %d)\n", l);
      return;
    }
  for (i = 2; i < len - 2; i += 2)
    checksum += get16 (buf, i);
  checksum &= 0xffff;
  tape_checksum = get16 (buf, len - 2);
  if (checksum != tape_checksum)
    fprintf (f, "computed checksum %04x, tape checksum %04x\n", checksum,
	     tape_checksum);
}


int main (int argc, char *argv[])
{
  uint32_t file = 0;
  uint32_t record = 0;
  uint32_t filebytes = 0;
  uint32_t tapebytes = 0;
  uint32_t len;
  char *srcfn = NULL;
  tape_handle_t src = NULL;
  uint8_t *buf;

  progname = argv [0];

  while (++argv, --argc)
    {
      if ((argv [0][0] == '-') && (argv [0][1] != '\0'))
	{
	  fatal (1, "unrecognized option '%s'\n", argv [0]);
	}
      else if (! srcfn)
	srcfn = argv [0];
      else
	fatal (1, NULL);
    }

  if (! srcfn)
    fatal (1, NULL);

  buf = (uint8_t *) malloc (MAX_REC_LEN);
  if (! buf)
    fatal (2, "can't allocate buffer\n");

  src = opentape (srcfn, 0, 0);
  if (! src)
    fatal (3, "can't open source tape\n");

  for (;;)
    {
      len = getrec (src, buf, MAX_REC_LEN);
      if (len == 0)
	{
	  printf ("total length of file %d = %d records, %d bytes\n",
		  file, record, filebytes);
	  dump_hp_2000_dump_tape_eof (stdout, file, record);
	  tapebytes += filebytes;
	  file++;
	  record = 0;
	  filebytes = 0;
	  printf ("start of file %d\n", file);
	  fflush (stdout);
	}
      else
	{
	  printf ("file %d record %d: length %d\n", file, record, len);
	  dump_hp_2000_dump_tape (stdout, file, record, buf, len);
	  /* dump (stdout, buf, len); */
	  fflush (stdout);
	  filebytes += len;
	  record++;
	}
    }

  closetape (src);

  return (0);
}
